@@ -1,3 +1,5 @@
+# frozen_string_literal: true
+
 # dbus.rb - Module containing the low-level D-Bus implementation
 #
 # This file is part of the ruby-dbus project
@@ -8,16 +10,15 @@
 # License, version 2.1 as published by the Free Software Foundation.
 # See the file "COPYING" for the exact licensing terms.
 
-require 'socket'
-require 'thread'
-require 'singleton'
+require "socket"
+require "singleton"
 
 # = D-Bus main module
 #
 # Module containing all the D-Bus modules and classes.
 module DBus
   # This represents a remote service. It should not be instantiated directly
-  # Use Bus::service()
+  # Use {Connection#service}
   class Service
     # The service name.
     attr_reader :name
@@ -28,7 +29,8 @@ module DBus
 
     # Create a new service with a given _name_ on a given _bus_.
     def initialize(name, bus)
-      @name, @bus = name, bus
+      @name = BusName.new(name)
+      @bus = bus
       @root = Node.new("/")
     end
 
@@ -40,57 +42,73 @@ module DBus
     # Perform an introspection on all the objects on the service
     # (starting recursively from the root).
     def introspect
-      if block_given?
-        raise NotImplementedError
-      else
-        rec_introspect(@root, "/")
-      end
+      raise NotImplementedError if block_given?
+
+      rec_introspect(@root, "/")
       self
     end
 
     # Retrieves an object at the given _path_.
+    # @param path [ObjectPath]
     # @return [ProxyObject]
-    def object(path)
-      node = get_node(path, true)
-      if node.object.nil?
-        node.object = ProxyObject.new(@bus, @name, path)
+    def [](path)
+      object(path, api: ApiOptions::A1)
+    end
+
+    # Retrieves an object at the given _path_
+    # whose methods always return an array.
+    # @param path [ObjectPath]
+    # @param api [ApiOptions]
+    # @return [ProxyObject]
+    def object(path, api: ApiOptions::A0)
+      node = get_node(path, create: true)
+      if node.object.nil? || node.object.api != api
+        node.object = ProxyObject.new(
+          @bus, @name, path,
+          api: api
+        )
       end
       node.object
     end
 
-    # Export an object _obj_ (an DBus::Object subclass instance).
+    # Export an object
+    # @param obj [DBus::Object]
     def export(obj)
       obj.service = self
-      get_node(obj.path, true).object = obj
+      get_node(obj.path, create: true).object = obj
     end
 
     # Undo exporting an object _obj_.
     # Raises ArgumentError if it is not a DBus::Object.
     # Returns the object, or false if _obj_ was not exported.
+    # @param obj [DBus::Object]
     def unexport(obj)
-      raise ArgumentError.new("DBus::Service#unexport() expects a DBus::Object argument") unless obj.kind_of?(DBus::Object)
+      raise ArgumentError, "DBus::Service#unexport() expects a DBus::Object argument" unless obj.is_a?(DBus::Object)
       return false unless obj.path
-      pathSep = obj.path.rindex("/") #last path seperator
-      parent_path = obj.path[1..pathSep-1]
-      node_name = obj.path[pathSep+1..-1]
 
-      parent_node = get_node(parent_path, false)
+      last_path_separator_idx = obj.path.rindex("/")
+      parent_path = obj.path[1..last_path_separator_idx - 1]
+      node_name = obj.path[last_path_separator_idx + 1..-1]
+
+      parent_node = get_node(parent_path, create: false)
       return false unless parent_node
+
       obj.service = nil
       parent_node.delete(node_name).object
     end
 
-    # Get the object node corresponding to the given _path_. if _create_ is
-    # true, the the nodes in the path are created if they do not already exist.
-    def get_node(path, create = false)
+    # Get the object node corresponding to the given *path*.
+    # @param path [ObjectPath]
+    # @param create [Boolean] if true, the the {Node}s in the path are created
+    #   if they do not already exist.
+    # @return [Node,nil]
+    def get_node(path, create: false)
       n = @root
-      path.sub(/^\//, "").split("/").each do |elem|
-        if not n[elem]
-          if not create
-            return nil
-          else
-            n[elem] = Node.new(elem)
-          end
+      path.sub(%r{^/}, "").split("/").each do |elem|
+        if !(n[elem])
+          return nil if !create
+
+          n[elem] = Node.new(elem)
         end
         n = n[elem]
       end
@@ -101,7 +119,9 @@ module DBus
     end
 
     #########
+
     private
+
     #########
 
     # Perform a recursive retrospection on the given current _node_
@@ -111,16 +131,16 @@ module DBus
       intfs, subnodes = IntrospectXMLParser.new(xml).parse
       subnodes.each do |nodename|
         subnode = node[nodename] = Node.new(nodename)
-        if path == "/"
-          subpath = "/" + nodename
-        else
-          subpath = path + "/" + nodename
-        end
+        subpath = if path == "/"
+                    "/#{nodename}"
+                  else
+                    "#{path}/#{nodename}"
+                  end
         rec_introspect(subnode, subpath)
       end
-      if intfs.size > 0
-        node.object = ProxyObjectFactory.new(xml, @bus, @name, path).build
-      end
+      return if intfs.empty?
+
+      node.object = ProxyObjectFactory.new(xml, @bus, @name, path).build
     end
   end
 
@@ -128,36 +148,36 @@ module DBus
   #
   # Class representing a node on an object path.
   class Node < Hash
-    # The D-Bus object contained by the node.
+    # @return [DBus::Object,DBus::ProxyObject,nil]
+    #   The D-Bus object contained by the node.
     attr_accessor :object
+
     # The name of the node.
+    # @return [String] the last component of its object path, or "/"
     attr_reader :name
 
     # Create a new node with a given _name_.
     def initialize(name)
+      super()
       @name = name
       @object = nil
     end
 
     # Return an XML string representation of the node.
     # It is shallow, not recursing into subnodes
-    def to_xml
+    # @param node_opath [String]
+    def to_xml(node_opath)
       xml = '<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
 "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
-<node>
 '
-      self.each_pair do |k, v|
-        xml += "<node name=\"#{k}\" />"
+      xml += "<node name=\"#{node_opath}\">\n"
+      each_key do |k|
+        xml += "  <node name=\"#{k}\" />\n"
       end
-      if @object
-        @object.intfs.each_pair do |k, v|
-          xml += %{<interface name="#{v.name}">\n}
-          v.methods.each_value { |m| xml += m.to_xml }
-          v.signals.each_value { |m| xml += m.to_xml }
-          xml +="</interface>\n"
-        end
+      @object&.intfs&.each_value do |v|
+        xml += v.to_xml
       end
-      xml += '</node>'
+      xml += "</node>"
       xml
     end
 
@@ -170,12 +190,15 @@ module DBus
     # Return instance inspect information, used by Node#inspect.
     def sub_inspect
       s = ""
-      if not @object.nil?
-        s += "%x " % @object.object_id
+      if !@object.nil?
+        s += format("%x ", @object.object_id)
       end
-      s + "{" + keys.collect { |k| "#{k} => #{self[k].sub_inspect}" }.join(",") + "}"
+      contents_sub_inspect = keys
+                             .map { |k| "#{k} => #{self[k].sub_inspect}" }
+                             .join(",")
+      "#{s}{#{contents_sub_inspect}}"
     end
-  end # class Inspect
+  end
 
   # FIXME: rename Connection to Bus?
 
@@ -198,9 +221,17 @@ module DBus
     def initialize(path)
       @message_queue = MessageQueue.new(path)
       @unique_name = nil
-      @method_call_replies = Hash.new
-      @method_call_msgs = Hash.new
-      @signal_matchrules = Hash.new
+
+      # @return [Hash{Integer => Proc}]
+      #   key: message serial
+      #   value: block to be run when the reply to that message is received
+      @method_call_replies = {}
+
+      # @return [Hash{Integer => Message}]
+      #   for debugging only: messages for which a reply was not received yet;
+      #   key == value.serial
+      @method_call_msgs = {}
+      @signal_matchrules = {}
       @proxy = nil
       @object_root = Node.new("/")
     end
@@ -209,20 +240,20 @@ module DBus
     # but do not block on the queue.
     # Called by a main loop when something is available in the queue
     def dispatch_message_queue
-      while (msg = @message_queue.pop(:non_block)) # FIXME EOFError
+      while (msg = @message_queue.pop(blocking: false)) # FIXME: EOFError
         process(msg)
       end
     end
 
     # Tell a bus to register itself on the glib main loop
     def glibize
-      require 'glib2'
+      require "glib2"
       # Circumvent a ruby-glib bug
-      @channels ||= Array.new
+      @channels ||= []
 
       gio = GLib::IOChannel.new(@message_queue.socket.fileno)
       @channels << gio
-      gio.add_watch(GLib::IOChannel::IN) do |c, ch|
+      gio.add_watch(GLib::IOChannel::IN) do |_c, _ch|
         dispatch_message_queue
         true
       end
@@ -317,7 +348,7 @@ module DBus
   </interface>
 </node>
 '
-# This apostroph is for syntax highlighting editors confused by above xml: "
+    # This apostroph is for syntax highlighting editors confused by above xml: "
 
     # @api private
     # Send a _message_.
@@ -330,11 +361,9 @@ module DBus
       ret = nil
       if reply_handler.nil?
         send_sync(message) do |rmsg|
-          if rmsg.is_a?(Error)
-            raise rmsg
-          else
-            ret = rmsg.params
-          end
+          raise rmsg if rmsg.is_a?(Error)
+
+          ret = rmsg.params
         end
       else
         on_return(message) do |rmsg|
@@ -361,8 +390,8 @@ module DBus
         send_sync_or_async(m).first
       else
         send_sync_or_async(m) do |*args|
-          # TODO test async introspection, is it used at all?
-          args.shift            # forget the message, pass only the text
+          # TODO: test async introspection, is it used at all?
+          args.shift # forget the message, pass only the text
           reply_handler.call(*args)
           nil
         end
@@ -379,11 +408,11 @@ module DBus
     # The returned object is a ProxyObject that has methods you can call to
     # issue somme METHOD_CALL messages, and to setup to receive METHOD_RETURN
     def introspect(dest, path)
-      if not block_given?
+      if !block_given?
         # introspect in synchronous !
         data = introspect_data(dest, path)
         pof = DBus::ProxyObjectFactory.new(data, self, dest, path)
-        return pof.build
+        pof.build
       else
         introspect_data(dest, path) do |async_data|
           yield(DBus::ProxyObjectFactory.new(async_data, self, dest, path).build)
@@ -406,23 +435,26 @@ module DBus
       # method calls to be serviced arrive before the reply for RequestName
       # (Ticket#29).
       proxy.RequestName(name, NAME_FLAG_REPLACE_EXISTING) do |rmsg, r|
-        if rmsg.is_a?(Error)  # check and report errors first
-	  raise rmsg
-	elsif r != REQUEST_NAME_REPLY_PRIMARY_OWNER
-          raise NameRequestError
-	end
+        # check and report errors first
+        raise rmsg if rmsg.is_a?(Error)
+        raise NameRequestError unless r == REQUEST_NAME_REPLY_PRIMARY_OWNER
       end
       @service = Service.new(name, self)
       @service
     end
 
     # Set up a ProxyObject for the bus itself, since the bus is introspectable.
+    # @return [ProxyObject] that always returns an array
+    #   ({DBus::ApiOptions#proxy_method_returns_array})
     # Returns the object.
     def proxy
-      if @proxy == nil
+      if @proxy.nil?
         path = "/org/freedesktop/DBus"
         dest = "org.freedesktop.DBus"
-        pof = DBus::ProxyObjectFactory.new(DBUSXMLINTRO, self, dest, path)
+        pof = DBus::ProxyObjectFactory.new(
+          DBUSXMLINTRO, self, dest, path,
+          api: ApiOptions::A0
+        )
         @proxy = pof.build["org.freedesktop.DBus"]
       end
       @proxy
@@ -431,45 +463,56 @@ module DBus
     # @api private
     # Wait for a message to arrive. Return it once it is available.
     def wait_for_message
-      @message_queue.pop                 # FIXME EOFError
+      @message_queue.pop # FIXME: EOFError
     end
 
     # @api private
-    # Send a message _m_ on to the bus. This is done synchronously, thus
+    # Send a message _msg_ on to the bus. This is done synchronously, thus
     # the call will block until a reply message arrives.
-    def send_sync(m, &retc) # :yields: reply/return message
-      return if m.nil? #check if somethings wrong
-      @message_queue.push(m)
-      @method_call_msgs[m.serial] = m
-      @method_call_replies[m.serial] = retc
+    # @param msg [Message]
+    # @param retc [Proc] the reply handler
+    # @yieldparam rmsg [MethodReturnMessage] the reply
+    # @yieldreturn [Array<Object>] the reply (out) parameters
+    def send_sync(msg, &retc) # :yields: reply/return message
+      return if msg.nil? # check if somethings wrong
+
+      @message_queue.push(msg)
+      @method_call_msgs[msg.serial] = msg
+      @method_call_replies[msg.serial] = retc
 
       retm = wait_for_message
-      return if retm.nil? #check if somethings wrong
-      
+      return if retm.nil? # check if somethings wrong
+
       process(retm)
-      while @method_call_replies.has_key? m.serial
+      while @method_call_replies.key? msg.serial
         retm = wait_for_message
         process(retm)
       end
+    rescue EOFError
+      new_err = DBus::Error.new("Connection dropped after we sent #{msg.inspect}")
+      raise new_err
     end
 
     # @api private
     # Specify a code block that has to be executed when a reply for
-    # message _m_ is received.
-    def on_return(m, &retc)
+    # message _msg_ is received.
+    # @param msg [Message]
+    def on_return(msg, &retc)
       # Have a better exception here
-      if m.message_type != Message::METHOD_CALL
+      if msg.message_type != Message::METHOD_CALL
         raise "on_return should only get method_calls"
       end
-      @method_call_msgs[m.serial] = m
-      @method_call_replies[m.serial] = retc
+
+      @method_call_msgs[msg.serial] = msg
+      @method_call_replies[msg.serial] = retc
     end
 
     # Asks bus to send us messages matching mr, and execute slot when
     # received
-    def add_match(mr, &slot)
+    # @param match_rule [MatchRule,#to_s]
+    def add_match(match_rule, &slot)
       # check this is a signal.
-      mrs = mr.to_s
+      mrs = match_rule.to_s
       DBus.logger.debug "#{@signal_matchrules.size} rules, adding #{mrs.inspect}"
       # don't ask for the same match if we override it
       unless @signal_matchrules.key?(mrs)
@@ -479,68 +522,77 @@ module DBus
       @signal_matchrules[mrs] = slot
     end
 
-    def remove_match(mr)
-      mrs = mr.to_s
-      unless @signal_matchrules.delete(mrs).nil?
-        # don't remove nonexisting matches.
-        # FIXME if we do try, the Error.MatchRuleNotFound is *not* raised
-        # and instead is reported as "no return code for nil"
-        proxy.RemoveMatch(mrs)
-      end
+    # @param match_rule [MatchRule,#to_s]
+    def remove_match(match_rule)
+      mrs = match_rule.to_s
+      rule_existed = @signal_matchrules.delete(mrs).nil?
+      # don't remove nonexisting matches.
+      return if rule_existed
+
+      # FIXME: if we do try, the Error.MatchRuleNotFound is *not* raised
+      # and instead is reported as "no return code for nil"
+      proxy.RemoveMatch(mrs)
     end
 
     # @api private
-    # Process a message _m_ based on its type.
-    def process(m)
-      return if m.nil? #check if somethings wrong
-      case m.message_type
+    # Process a message _msg_ based on its type.
+    # @param msg [Message]
+    def process(msg)
+      return if msg.nil? # check if somethings wrong
+
+      case msg.message_type
       when Message::ERROR, Message::METHOD_RETURN
-        raise InvalidPacketException if m.reply_serial == nil
-        mcs = @method_call_replies[m.reply_serial]
-        if not mcs
-          DBus.logger.debug "no return code for mcs: #{mcs.inspect} m: #{m.inspect}"
+        raise InvalidPacketException if msg.reply_serial.nil?
+
+        mcs = @method_call_replies[msg.reply_serial]
+        if !mcs
+          DBus.logger.debug "no return code for mcs: #{mcs.inspect} msg: #{msg.inspect}"
         else
-          if m.message_type == Message::ERROR
-            mcs.call(Error.new(m))
+          if msg.message_type == Message::ERROR
+            mcs.call(Error.new(msg))
           else
-            mcs.call(m)
+            mcs.call(msg)
           end
-          @method_call_replies.delete(m.reply_serial)
-          @method_call_msgs.delete(m.reply_serial)
+          @method_call_replies.delete(msg.reply_serial)
+          @method_call_msgs.delete(msg.reply_serial)
         end
       when DBus::Message::METHOD_CALL
-        if m.path == "/org/freedesktop/DBus"
+        if msg.path == "/org/freedesktop/DBus"
           DBus.logger.debug "Got method call on /org/freedesktop/DBus"
         end
-        node = @service.get_node(m.path)
-        if not node
-          reply = Message.error(m, "org.freedesktop.DBus.Error.UnknownObject",
-                                "Object #{m.path} doesn't exist")
+        node = @service.get_node(msg.path, create: false)
+        if !node
+          reply = Message.error(msg, "org.freedesktop.DBus.Error.UnknownObject",
+                                "Object #{msg.path} doesn't exist")
           @message_queue.push(reply)
         # handle introspectable as an exception:
-        elsif m.interface == "org.freedesktop.DBus.Introspectable" and
-            m.member == "Introspect"
-          reply = Message.new(Message::METHOD_RETURN).reply_to(m)
+        elsif msg.interface == "org.freedesktop.DBus.Introspectable" &&
+              msg.member == "Introspect"
+          reply = Message.new(Message::METHOD_RETURN).reply_to(msg)
           reply.sender = @unique_name
-          reply.add_param(Type::STRING, node.to_xml)
+          xml = node.to_xml(msg.path)
+          reply.add_param(Type::STRING, xml)
           @message_queue.push(reply)
         else
           obj = node.object
-          return if obj.nil?    # FIXME, pushes no reply
-          obj.dispatch(m) if obj
+          return if obj.nil? # FIXME, pushes no reply
+
+          obj&.dispatch(msg)
         end
       when DBus::Message::SIGNAL
         # the signal can match multiple different rules
-        @signal_matchrules.each do |mrs, slot|
-          if DBus::MatchRule.new.from_s(mrs).match(m)
-            slot.call(m)
+        # clone to allow new signale handlers to be registered
+        @signal_matchrules.dup.each do |mrs, slot|
+          if DBus::MatchRule.new.from_s(mrs).match(msg)
+            slot.call(msg)
           end
         end
       else
-        DBus.logger.debug "Unknown message type: #{m.message_type}"
+        # spec(Message Format): Unknown types must be ignored.
+        DBus.logger.debug "Unknown message type: #{msg.message_type}"
       end
-    rescue Exception => ex
-      raise m.annotate_exception(ex)
+    rescue Exception => e
+      raise msg.annotate_exception(e)
     end
 
     # Retrieves the Service with the given _name_.
@@ -550,11 +602,16 @@ module DBus
       # anything
       Service.new(name, self)
     end
-    alias :[] :service
+    alias [] service
 
     # @api private
     # Emit a signal event for the given _service_, object _obj_, interface
     # _intf_ and signal _sig_ with arguments _args_.
+    # @param service [Service]
+    # @param obj [DBus::Object]
+    # @param intf [Interface]
+    # @param sig [Signal]
+    # @param args arguments for the signal
     def emit(service, obj, intf, sig, *args)
       m = Message.new(DBus::Message::SIGNAL)
       m.path = obj.path
@@ -585,8 +642,7 @@ module DBus
       end
       @service = Service.new(@unique_name, self)
     end
-  end # class Connection
-
+  end
 
   # = D-Bus session bus class
   #
@@ -610,18 +666,20 @@ module DBus
     def self.address_from_file
       # systemd uses /etc/machine-id
       # traditional dbus uses /var/lib/dbus/machine-id
-      machine_id_path = Dir['{/etc,/var/lib/dbus,/var/db/dbus}/machine-id'].first
+      machine_id_path = Dir["{/etc,/var/lib/dbus,/var/db/dbus}/machine-id"].first
       return nil unless machine_id_path
+
       machine_id = File.read(machine_id_path).chomp
 
       display = ENV["DISPLAY"][/:(\d+)\.?/, 1]
 
       bus_file_path = File.join(ENV["HOME"], "/.dbus/session-bus/#{machine_id}-#{display}")
-      return nil unless File.exists?(bus_file_path)
+      return nil unless File.exist?(bus_file_path)
 
       File.open(bus_file_path).each_line do |line|
         if line =~ /^DBUS_SESSION_BUS_ADDRESS=(.*)/
-          return $1
+          address = Regexp.last_match(1)
+          return address[/\A'(.*)'\z/, 1] || address[/\A"(.*)"\z/, 1] || address
         end
       end
     end
@@ -632,7 +690,6 @@ module DBus
     include Singleton
   end
 
-
   # = D-Bus system bus class
   #
   # The system bus is a system-wide bus mostly used for global or
@@ -643,26 +700,25 @@ module DBus
   class ASystemBus < Connection
     # Get the default system bus.
     def initialize
-      super(SystemSocketName)
+      super(SYSTEM_BUS_ADDRESS)
       send_hello
     end
   end
-  
+
   # = D-Bus remote (TCP) bus class
   #
-  # This class may be used when connecting to remote (listening on a TCP socket) 
+  # This class may be used when connecting to remote (listening on a TCP socket)
   # busses. You can also use it to connect to other non-standard path busses.
-  # 
+  #
   # The specified socket_name should look like this:
   # (for TCP)         tcp:host=127.0.0.1,port=2687
   # (for Unix-socket) unix:path=/tmp/my_funky_bus_socket
-  # 
-  # you'll need to take care about authentification then, more info here: 
-  # http://github.com/pangdudu/ruby-dbus/blob/master/README.rdoc
+  #
+  # you'll need to take care about authentification then, more info here:
+  # https://gitlab.com/pangdudu/ruby-dbus/-/blob/master/README.rdoc
   class RemoteBus < Connection
-
     # Get the remote bus.
-    def initialize socket_name
+    def initialize(socket_name)
       super(socket_name)
       send_hello
     end
@@ -675,13 +731,13 @@ module DBus
 
   # Shortcut for the {SystemBus} instance
   # @return [Connection]
-  def DBus.system_bus
+  def self.system_bus
     SystemBus.instance
   end
 
   # Shortcut for the {SessionBus} instance
   # @return [Connection]
-  def DBus.session_bus
+  def self.session_bus
     SessionBus.instance
   end
 
@@ -693,7 +749,7 @@ module DBus
   class Main
     # Create a new main event loop.
     def initialize
-      @buses = Hash.new
+      @buses = {}
       @quitting = false
     end
 
@@ -712,13 +768,14 @@ module DBus
       # before blocking, empty the buffers
       # https://bugzilla.novell.com/show_bug.cgi?id=537401
       @buses.each_value do |b|
-        while m = b.message_queue.message_from_buffer_nonblock
+        while (m = b.message_queue.message_from_buffer_nonblock)
           b.process(m)
         end
       end
-      while not @quitting and not @buses.empty?
+      while !@quitting && !@buses.empty?
         ready = IO.select(@buses.keys, [], [], 5) # timeout 5 seconds
         next unless ready # timeout exceeds so continue unless quitting
+
         ready.first.each do |socket|
           b = @buses[socket]
           begin
@@ -727,11 +784,11 @@ module DBus
             @buses.delete socket # this bus died
             next
           end
-          while m = b.message_queue.message_from_buffer_nonblock
+          while (m = b.message_queue.message_from_buffer_nonblock)
             b.process(m)
           end
         end
       end
     end
-  end # class Main
-end # module DBus
+  end
+end
