@@ -1,4 +1,6 @@
 #!/usr/bin/env rspec
+# frozen_string_literal: true
+
 # Test thread safety
 require_relative "spec_helper"
 require "dbus"
@@ -15,7 +17,6 @@ describe "ThreadSafetyTest" do
         bus = DBus::ASessionBus.new
         svc = bus.service("org.ruby.service")
         obj = svc.object("/org/ruby/MyInstance")
-        obj.introspect
         obj.default_iface = "org.ruby.SampleInterface"
 
         10.times do |i|
@@ -26,6 +27,6 @@ describe "ThreadSafetyTest" do
         end
       end
     end
-    jobs.each do |thread| thread.join end
+    jobs.each(&:join)
   end
 end
